library IEEE;
use IEEE.std_logic_1164.ALL;
use IEEE.std_logic_unsigned.ALL;
use IEEE.std_logic_arith.ALL;
use work.lapg_pack.ALL;

entity LAPG is
	generic(INPUT_BYTE : integer := 255;
		   OUTPUT_BYTE : integer := 255;
		   OSC : integer := 5
	);
	
	port(CLK : in std_logic;
	     RESET : in std_logic;
		RxD : in std_logic;
		Sampled_Data : in std_logic_vector((OUTPUT_BYTE * 8) - 1 downto 0);
		TxD : out std_logic;
		DCLK : out std_logic;
		Injection_Data	: out std_logic_vector((INPUT_BYTE * 8) - 1 downto 0)
	);
end LAPG;

architecture RTL of LAPG is
	constant MAX_BAUD_CNT 			: integer range 0 to 434 := (OSC * 10000 / MAX_BAUD_RATE);
	constant MAX_START_CNT 			: integer range 0 to 217 := MAX_BAUD_CNT / 2;
	constant MAX_PG_CNT				: integer range 0 to 108 := MAX_BAUD_CNT / 4;
	constant CONF_INPUT_LENGTH		: std_logic_vector(7 downto 0) := CONV_STD_LOGIC_VECTOR(INPUT_BYTE, 8);
	constant CONF_OUTPUT_LENGTH		: std_logic_vector(7 downto 0) := CONV_STD_LOGIC_VECTOR(OUTPUT_BYTE, 8);
	
	-- STATE
	signal C_STATE : STATE;
	signal N_STATE : STATE;

	-- DESIGN CLOCK
	signal Design_Clk : std_logic;

	-- Synchronous Count
	signal Sync_Cnt_En : std_logic;
	signal Sync_Cnt : integer range 0 to MAX_BAUD_CNT;
	
	-- Bit Count
	signal Bit_Cnt_En : std_logic_vector(1 downto 0);
	signal Bit_Cnt : integer range 0 to MAX_TRANS_CNT;
		
	-- Byte Count
	signal Byte_Cnt_En : std_logic_vector(1 downto 0);
	signal Byte_Cnt : integer range 0 to INPUT_BYTE + OUTPUT_BYTE;
	
	-- Toggling Count
	signal Toggling_Cnt_En : std_logic_vector(1 downto 0);
	signal Toggling_Cnt : integer range 0 to 16384;

	-- INPUT_BYTE*8-bit Receive Data Register & 8-bit Receive Command Register
	signal Rec_Data_En : std_logic;
	signal Rec_Data_Reg : std_logic_vector(INPUT_BYTE * 8 - 1 downto 0);
	signal Rec_Com_En : std_logic;
	signal Rec_Com_Reg : std_logic_vector(7 downto 0);

	-- 7-bit Exe Register		  
	signal Exe_Reg_Ld : std_logic;
	signal Exe_Reg : std_logic_vector(6 downto 0);

	-- 6-bit Reset Register
	signal Reset_Reg_Ld : std_logic;
	signal Reset_Reg : std_logic_vector(5 downto 0);

	-- 3-bit Transmit Select
	signal Trans_Sel_En : std_logic;
	signal Trans_Sel : std_logic_vector(2 downto 0);
	signal Trans_Sel_Buf : std_logic_vector(2 downto 0);
  	
	-- Design_Clock Toggling Signal
	signal Dclk_Toggle : std_logic_vector(1 downto 0);
	
	-- Trans Data Register
	signal Trans_Data : std_logic_vector(11 downto 0);
	signal Trans_Data_En : std_logic;

	-- Injection Enable Signal
	signal Injection_En : std_logic;
										 
	-- Data Mode Signal
	signal Data_Mode : std_logic := '0';
	signal Data_Mode_En : std_logic_vector(1 downto 0);
	
	-- Toggling Counter
	signal MAX_TOGGLING_CNT : integer range 0 to 16384 := 0;
	
	-- Configuration Check Data
	signal Config_Data : std_logic_vector(15 downto 0);	

	-- Propagation Delay Counter
	signal MAX_DELAY_CNT : integer range 0 to 972 := MAX_PG_CNT;
begin
	DCLK <= Design_Clk;

	-- TRANSMIT
	TxD <= Trans_Data(Bit_Cnt) when C_STATE = TRANSMIT else '1'; 

	-- INJECTION
	Injection_Data <= Rec_Data_Reg when Injection_En = '1' else
				   (others => '0');
	
	-- ASSIGN CONFIGURATION_DATA
	Config_Data <= CONF_OUTPUT_LENGTH & CONF_INPUT_LENGTH;

	-- COMPUTATION TOGGLING_COUNT
	MAX_TOGGLING_CNT <= cal_count(Exe_Reg(4), Exe_Reg(3 downto 0));
	
	-- ASSIGN STATE
	SEQ : process(CLK, RESET) 		  
	begin
		if(RESET = '1') then
			C_STATE <= REC_READY;
		elsif(CLK = '1' and CLK'event) then
			C_STATE <= N_STATE;
		end if;
	end process;

	-- MAIN FSM
	COMB : process(C_STATE, RxD, Sync_Cnt, Bit_Cnt, Byte_Cnt, Data_Mode, 
				Rec_Data_Reg, Rec_Com_Reg, MAX_TOGGLING_CNT, MAX_DELAY_CNT, 
				Exe_Reg, Reset_Reg, Toggling_Cnt, Trans_Sel 
				)
	begin
		Bit_Cnt_En <= "00";
		Byte_Cnt_En <= "00";
		Toggling_Cnt_En <= "00";
		Sync_Cnt_En <= '0';
		Rec_Data_En <= '0';
		Rec_Com_En <= '0';
		Exe_Reg_Ld <= '0';
		Reset_Reg_Ld <= '0';
		Trans_Data_En <= '0';
		Dclk_Toggle <= "00";
		Data_Mode_En <= "00";
		Trans_Sel_Buf <= "000";
		Trans_Sel_En <= '0';
		Injection_En <= '0';
							
		case C_STATE is
			when REC_READY =>
				Byte_Cnt_En <= "11";
				Dclk_Toggle <= "11";
				Toggling_Cnt_En <= "11";
				Data_Mode_En <= "11";
				if(RxD = '0') then
					if(Sync_Cnt = MAX_START_CNT) then
						N_STATE <= RECEIVE;
					else
						Sync_Cnt_En <= '1';
						N_STATE <= REC_READY;
					end if;
				else
					N_STATE <= REC_READY;
					Injection_En <= '1';
				end if;
			when RECEIVE =>
				Dclk_Toggle <= "11";
				Toggling_Cnt_En <= "11";
				Data_Mode_En <= "11";
				if(Sync_Cnt = MAX_BAUD_CNT) then
					if(Data_Mode = '1') then
						case Exe_Reg(6) is
							when '0' =>
								if(Bit_Cnt = MAX_BIT_CNT) then
									N_STATE <= PG_DELAY;
								else
									Bit_Cnt_En <= "01";
									N_STATE <= RECEIVE;
								end if;
							when others =>
								if(Bit_Cnt = MAX_BIT_CNT) then
									if(Byte_Cnt = INPUT_BYTE - 1) then
										N_STATE <= PG_DELAY;
										Injection_En <= '1';
									else
										Byte_Cnt_En <= "01";
										N_STATE <= REC_READY;
									end if;
								else
									Rec_Data_En <= '1';
									Bit_Cnt_En <= "01";
									Byte_Cnt_En <= "11";
									N_STATE <= RECEIVE;
								end if;
						end case;
					else
						
						if(Bit_Cnt = MAX_BIT_CNT) then
							N_STATE <= COMMAND;
						else
							Rec_Com_En <= '1';
							Bit_Cnt_En <= "01";
							N_STATE <= RECEIVE;
						end if;
					end if;
				else
					Sync_Cnt_En <= '1';
					Bit_Cnt_En <= "11";
					Byte_Cnt_En <= "11";
					N_STATE <= RECEIVE;
				end if;
			when COMMAND =>
				Data_Mode_En <= "11";
				Toggling_Cnt_En <= "11";
				Dclk_Toggle <= "11";
				
				case Rec_Com_Reg(7 downto 6) is
					when "10" =>
						Reset_Reg_Ld <= '1';
						Trans_Sel_Buf <= "010";
						Trans_Sel_En <= '1';
					when "11" =>
						Reset_Reg_Ld <= '1';
						Trans_Sel_Buf <= "001";
						Trans_Sel_En <= '1';
					when others =>
						Exe_Reg_Ld <= '1';
						Trans_Sel_Buf <= "011";
						Trans_Sel_En <= '1';
				end case;
				N_STATE <= TRANS_READY;

			when PG_DELAY =>
				Data_Mode_En <= "11";
				Dclk_Toggle <= "11";
				Toggling_Cnt_En <= "11";
				Injection_En <= '1';
				if(Sync_Cnt = MAX_DELAY_CNT) then
					Trans_Sel_En <= '1';
					Trans_Sel_Buf <= "100";
				     case Exe_Reg(5) is
						when '0' =>
							N_STATE <= TOGGLING;
							case Reset_Reg(3) is
								when '0' =>
									Dclk_Toggle <= "10";
								when others =>
									Dclk_Toggle <= "11";
							end case;
						when others =>
							N_STATE <= TRANS_READY;
					end case;
				else
					Sync_Cnt_En <= '1';
					N_STATE <= PG_DELAY;
				end if;
			when TOGGLING =>
				Data_Mode_En <= "11";
				Injection_En <= '1';
			  	Dclk_Toggle <= "10";
				if(Toggling_Cnt = MAX_TOGGLING_CNT - 1) then
					N_STATE <= REC_READY;
					Data_Mode_En <= "00";
				else
					Toggling_Cnt_En <= "01";
					N_STATE <= REC_READY;
					case Reset_Reg(2) is
						when '1' =>
							case Exe_Reg(6) is
								when '0' =>
									N_STATE <= PG_DELAY;
								when others =>
									N_STATE <= REC_READY;
							end case;
						when others =>
							N_STATE <= REC_READY;
					end case;
				end if;
			when TRANS_READY =>
				Toggling_Cnt_En <= "11";
				Data_Mode_En <= "11";
				Dclk_Toggle <= "11";
				Injection_En <= '1';
				case Trans_Sel is
					when "001" =>
						if(Byte_Cnt = MAX_CONF_BYTE_NUM) then
							N_STATE <= REC_READY;
							Byte_Cnt_En <= "00";
						else
							Byte_Cnt_En <= "01";
							Trans_Data_En <= '1';
							N_STATE <= TRANSMIT;				
						end if;
					when "010" =>
						Trans_Data_En <= '1';
						N_STATE <= TRANSMIT;
						Dclk_Toggle <= "01";
					when "011" =>
						Trans_Data_En <= '1';
						N_STATE <= TRANSMIT;
					when "100" =>
						if(Byte_Cnt = OUTPUT_BYTE) then
							N_STATE <= TOGGLING;
							Byte_Cnt_En <= "00";
							if(Reset_Reg(3) = '0') then
								Dclk_Toggle <= "10";
							end if;
						else
							Byte_Cnt_En <= "01";
							Trans_Data_En <= '1';
							N_STATE <= TRANSMIT;
						end if;
					when others =>
						N_STATE <= TRANSMIT;
						Trans_Data_En <= '1';
				end case;

			when TRANSMIT =>
				Toggling_Cnt_En <= "11";
				Data_Mode_En <= "11";
				Dclk_Toggle <= "11";
				Byte_Cnt_En <= "11";
				Bit_Cnt_En <= "11";
				Injection_En <= '1';
				if(Sync_Cnt = MAX_BAUD_CNT) then
					case Trans_Sel is
						when "001" =>
							if(Bit_Cnt = MAX_TRANS_CNT) then
								N_STATE <= TRANS_READY;
							else
								Bit_Cnt_En <= "01";
								N_STATE <= TRANSMIT;
							end if;
						when "100" =>
							if(Bit_Cnt = MAX_TRANS_CNT) then
								N_STATE <= TRANS_READY;
							else
								Bit_Cnt_En <= "01";
								N_STATE <= TRANSMIT;
							end if;
						when "011" =>
							if(Bit_Cnt = MAX_TRANS_CNT) then
								case Reset_Reg(2) is
									when '1' =>
										case Exe_Reg(6) is
											when '0' =>
												N_STATE <= PG_DELAY;
											when others =>
												N_STATE <= REC_READY;
												Data_Mode_En <= "01";
										end case;							
									when others =>
										N_STATE <= REC_READY;
										Data_Mode_En <= "01";
								end case;
							else
								Bit_Cnt_En <= "01";
								N_STATE <= TRANSMIT;
							end if;
						when others =>
							if(Bit_Cnt = MAX_TRANS_CNT) then
								N_STATE <= REC_READY;
								Bit_Cnt_En <= "00";
							else
								Bit_Cnt_En <= "01";
								N_STATE <= TRANSMIT;
							end if;
					end case;
				else
					Sync_Cnt_En <= '1';
					Bit_Cnt_En <= "11";
					N_STATE <= TRANSMIT;
				end if;
			
			when others =>
				N_STATE <= REC_READY;
		end case;

	end process;

	-- Design Clock Generate
	DCLK_GENERATE : process(CLK, RESET)
	begin
		if(RESET = '1') then
			design_clk <= '0';
		elsif(CLK = '1' and CLK'event) then
			case Dclk_Toggle is
				when "00" =>
					Design_Clk <= '0';
				when "01" =>
					case Reset_Reg(5 downto 4) is
						when "00" =>
							Design_Clk <= Design_Clk;
						when "01" =>
							Design_Clk <= '1';
						when "10" =>
							Design_Clk <= '0';
						when others =>
							Design_Clk <= '0';
					end case;
				when "10" =>
					Design_Clk <= not Design_Clk;
				when others =>
					Design_Clk <= Design_Clk;
			end case;
		end if;
	end process;

	-- Synchronous Count Generate
	SYNC_COUNT_GENERATE : process(CLK, RESET)
	begin
		if(RESET = '1') then
			Sync_Cnt <= 0;
		elsif(CLK = '1' and CLK'event) then
			if(Sync_Cnt_En = '1') then
				Sync_Cnt <= Sync_Cnt + 1;
			else
				Sync_Cnt <= 0;
			end if;
		end if;
	end process;

	-- Bit Count Generate
	BIT_COUNT_GENERATE : process(CLK, RESET)
	begin
		if(RESET = '1') then
			Bit_Cnt <= 0;
		elsif(CLK = '1' and CLK'event) then
			case Bit_Cnt_En is
				when "00" =>
					Bit_Cnt <= 0;
				when "01" =>
					Bit_Cnt <= Bit_Cnt + 1;
				when others =>
					Bit_Cnt <= Bit_Cnt;
			end case;
		end if;
	end process;

	-- Byte Count Generate
	BYTE_COUNT_GENERATE : process(CLK, RESET)
	begin
		if(RESET = '1') then
			Byte_Cnt <= 0;
		elsif(CLK = '1' and CLK'event) then
			case Byte_Cnt_En is
				when "00" =>
					Byte_Cnt <= 0;
				when "01" =>
					Byte_Cnt <= Byte_Cnt + 1;
				when others =>
					Byte_Cnt <= Byte_Cnt;
			end case;
		end if;
	end process;

	-- Toggling Count Generate
	TOGGLING_COUNT_GENERATE : process(CLK, RESET)
	begin
		if(RESET = '1') then
			Toggling_Cnt <= 0;
		elsif(CLK = '1' and CLK'event) then
			case Toggling_Cnt_En is
				when "00" =>
					Toggling_Cnt <= 0;
				when "01" =>
					Toggling_Cnt <= Toggling_Cnt + 1;
				when "11" =>
					Toggling_Cnt <= Toggling_Cnt;
				when others =>
					Toggling_Cnt <= 0;
			end case;
		end if;
	end process;

	-- Receive Data Register
	REC_DATA_REGISTER : process(CLK, RESET)
	begin
		if(RESET = '1') then
			Rec_Data_Reg <= (others => '0');
		elsif(CLK = '1' and CLK'event) then
			if(Rec_Data_En = '1') then
				Rec_Data_Reg <= RxD & Rec_Data_Reg(INPUT_BYTE * 8 - 1 downto 1);
			else
				Rec_Data_Reg <= Rec_Data_Reg;
			end if;
		end if;
	end process;

	-- Receive Command Register
	REC_COMMAND_REGISTER : process(CLK, RESET)
	begin
		if(RESET = '1') then
			Rec_Com_Reg <= (others => '0');
		elsif(CLK = '1' and CLK'event) then
			if(Rec_Com_En = '1') then
				Rec_Com_Reg <= RxD & Rec_Com_Reg(7 downto 1);
			else
				Rec_Com_Reg <= Rec_Com_Reg;
			end if;
		end if;
	end process;
	
	-- 8-bit Exe Register
	EXE_REGISTER : process(CLK, RESET)
	begin
		if(RESET = '1') then
			Exe_Reg <= (others => '0');
		elsif(CLK = '1' and CLK'event) then
			if(Exe_Reg_Ld = '1') then
				Exe_Reg <= Rec_Com_Reg(6 downto 0);
			else
				Exe_Reg <= Exe_Reg;
			end if;
		end if;
	end process;

	-- 8-bit Reset Register
	RESET_REGISTER : process(CLK, RESET)
	begin
		if(RESET = '1') then
			Reset_Reg <= (others => '0');
		elsif(CLK = '1' and CLK'event) then
			if(Reset_Reg_Ld = '1') then
				Reset_Reg <= Rec_Com_Reg(5 downto 0);
			else
				Reset_Reg <= Reset_Reg;
			end if;
		end if;
	end process;

	-- Propagation_Delay Generate
	PROPAGATION_DELAY_GENERATE : process(CLK, RESET)
	begin
		if(RESET = '1') then
			MAX_DELAY_CNT <= 0;
		elsif(CLK'event and CLK = '1') then
			case Rec_Com_Reg(1 downto 0) is
				when "00" =>
					MAX_DELAY_CNT <= MAX_DELAY_CNT;
				when "01" =>
					MAX_DELAY_CNT <= MAX_DELAY_CNT * 2;
					if(MAX_DELAY_CNT >= 972) then
						MAX_DELAY_CNT <= 972;
					end if;
				when "10" =>
					if(MAX_DELAY_CNT <= MAX_PG_CNT) then
						MAX_DELAY_CNT <= MAX_PG_CNT;
					else
						MAX_DELAY_CNT <= MAX_DELAY_CNT / 2;
					end if;
				when others =>
					MAX_DELAY_CNT <= MAX_PG_CNT;
			end case;
		end if;
	end process;
	
	-- Data Mode Generate
	DATA_MODE_GENERATE : process(CLK, RESET)
	begin
		if(RESET = '1') then
			Data_Mode <= '0';
		elsif(CLK = '1' and CLK'event) then
			case Data_Mode_En is
				when "00" =>
					Data_Mode <= '0';
				when "01" =>
					Data_Mode <= '1';
				when others =>
					Data_Mode <= Data_Mode;
			end case;
		end if;
	end process;
	
	-- Trans Data Generate
	TRANS_DATA_GENERATE : process(CLK, RESET)
	begin
		if(RESET = '1') then
			Trans_Data <= (others => '0');
		elsif(CLK = '1' and CLK'event) then
			if(Trans_Data_En = '1') then
				case Trans_Sel is
					when "001" =>
--						Trans_Data <= "11" & Config_Data((Byte_Cnt * 8 + 7) downto (Byte_Cnt * 8 + 0)) & "01";
						Trans_Data <= "11" & Config_Data(Byte_Cnt * 8 + 7) & Config_Data(Byte_Cnt * 8 + 6) 
						& Config_Data(Byte_Cnt * 8 + 5) & Config_Data(Byte_Cnt * 8 + 4) 
						& Config_Data(Byte_Cnt * 8 + 3) & Config_Data(Byte_Cnt * 8 + 2) 
						& Config_Data(Byte_Cnt * 8 + 1) & Config_Data(Byte_Cnt * 8) & "01";
--						Trans_Data <= "11" & "00000000" & "01";
					when "010" =>
						Trans_Data <= "11" & ACK_OK & "01";
					when "011" =>
						Trans_Data <= "11" & ACK_OK & "01";
					when "100" =>
--						Trans_Data <= "11" & Sampled_Data((Byte_Cnt * 8 + 7) downto (Byte_Cnt * 8)) & "01";
						Trans_Data <= "11" & Sampled_Data(Byte_Cnt * 8 + 7) & Sampled_Data(Byte_Cnt * 8 + 6)
						 & Sampled_Data(Byte_Cnt * 8 + 5) & Sampled_Data(Byte_Cnt * 8 + 4)
						 & Sampled_Data(Byte_Cnt * 8 + 3) & Sampled_Data(Byte_Cnt * 8 + 2)
						 & Sampled_Data(Byte_Cnt * 8 + 1) & Sampled_Data(Byte_Cnt * 8) & "01";
--						Trans_Data <= "11" & "00000000" & "01";
					when others =>
						Trans_Data <= "11" & "11111111" & "01";
				end case;
			else
				Trans_Data <= Trans_Data;
			end if;
		end if;
	end process;

	TRANS_SEL_BUFFER : process(CLK, RESET)
	begin
		if(RESET = '1') then
			Trans_Sel <= (others => '0');
		elsif(CLK = '1' and CLK'event) then
			if(Trans_Sel_En = '1') then
				Trans_Sel <= Trans_Sel_Buf;
			else
				Trans_Sel <= Trans_Sel;
			end if;
		end if;
	end process;
end RTL;